# coding: utf-8
# pylint: disable=protected-access
from types import NoneType

from django.core.exceptions import FieldDoesNotExist
from django.core.exceptions import ValidationError
from django.db import models
from django.db.models.fields.related import RelatedField
from django.db.models.fields.reverse_related import ForeignObjectRel
from django.db.models.fields.reverse_related import OneToOneRel
from django.utils.encoding import force_unicode
from m3_django_compat import ModelOptions

from educommon.django.db.utils import LazyModel
from educommon.utils.misc import cached_property

from . import constants


def skip_field(model, field):
    u"""Возвращает True, если поле модели должно игнорироваться конструктором.

    Игнорируются:
        - поля с типами ``AutoField`` и ``FileField``;
        - поля указанные в ``model.report_constructor_params['except']``;
        - поля *не* указанные в ``model.report_constructor_params['only']``;
        - поля моделей, у которых ``model.report_constructor_params['skip']``
          содержит истинное значение;

    :rtype: bool
    """
    if isinstance(field, (models.AutoField, models.FileField)):
        return True

    if not hasattr(model, 'report_constructor_params'):
        return False

    params = model.report_constructor_params
    assert isinstance(params, dict), params
    assert sum(
        1 if param in params else 0
        for param in ('skip', 'except', 'only')
    ) == 1, (u'Должен быть указан только один параметр.', model, params)

    if 'skip' in params and params['skip']:
        return True

    if isinstance(field, ForeignObjectRel):
        accessor_name = field.get_accessor_name()
    else:
        accessor_name = field.name

    if 'except' in params and accessor_name in params['except']:
        return True

    if 'only' in params and accessor_name not in params['only']:
        return True

    return False


class ColumnDescriptor(object):

    u"""Дескриптор поля модели, входящего в состав столбца отчета.

    Предназначен для извлечения из моделей необходимой информации.
    """

    def __init__(self, parent, field, level=0):
        u"""Инициализация экземпляра класса.

        :param parent: Дескриптор верхнего уровня. Определяет дескриптор для
            поля, из которого модель ссылается на поле данного дейскриптора.
        :type parent: ColumnDescriptor or NoneType

        :param field: Поле модели, соответствующее данному дескриптору.

        :param int level: Уровень вложенности дескриптора.
        """
        assert isinstance(parent, (ColumnDescriptor, NoneType)), parent

        self.parent = parent
        self.field = field
        self.level = level
        self.children = []

        if self.parent:
            self.parent.children.append(self)

    @staticmethod
    def create(model, field_name, parent=None, level=0):
        model_params = getattr(model, 'report_constructor_params', {})
        assert sum(
            1 if param in model_params else 0
            for param in ('skip', 'except', 'only')
        ) <= 1, (
            u'Может быть указан только один параметр.', model, model_params
        )

        attr_name, _, nested_attr = field_name.partition('.')

        for field in getattr(model, '_meta').get_fields():
            if skip_field(model, field):
                continue

            if isinstance(field, ForeignObjectRel):
                accessor_name = field.get_accessor_name()
                related_model = field.related_model
            else:
                accessor_name = field.name
                if isinstance(field, RelatedField):
                    related_model = field.rel.to
                else:
                    related_model = None

            if accessor_name == attr_name:
                column_field = field
                break
        else:
            raise FieldDoesNotExist(field_name)

        result = ColumnDescriptor(parent, column_field, level)
        if nested_attr:
            if not related_model:
                raise FieldDoesNotExist(field_name)

            try:
                result = ColumnDescriptor.create(
                    related_model, nested_attr, result, level + 1
                )
            except FieldDoesNotExist as error:
                raise FieldDoesNotExist('.'.join((attr_name, str(error))))

        return result

    @property
    def accessor_name(self):
        if isinstance(self.field, ForeignObjectRel):
            result = self.field.get_accessor_name()
        else:
            result = self.field.name

        return force_unicode(result)

    @property
    def lookup(self):
        return force_unicode(self.field.name)

    @property
    def full_accessor_name(self):
        if self.parent:
            return '.'.join(
                (self.parent.full_accessor_name, self.accessor_name)
            )
        else:
            return self.accessor_name

    @property
    def full_lookup(self):
        if self.parent:
            return '__'.join(
                (self.parent.full_lookup, self.lookup)
            )
        else:
            return self.lookup

    def __repr__(self, *args, **kwargs):
        return '{}<{}>'.format(
            self.__class__.__name__, self.full_accessor_name
        )

    @property
    def data_type(self):
        u"""Тип данных в столбце.

        :rtype: str
        """
        if self.field.concrete and self.field.choices:
            return constants.CT_CHOICES

        types_map = (
            (models.IntegerField, constants.CT_NUMBER),
            (models.FloatField, constants.CT_NUMBER),
            (models.DecimalField, constants.CT_NUMBER),
            (models.CharField, constants.CT_TEXT),
            (models.TextField, constants.CT_TEXT),
            (models.DateField, constants.CT_DATE),
            (models.TimeField, constants.CT_TIME),
            (models.DateTimeField, constants.CT_DATETIME),
            (models.ForeignKey, constants.CT_DIRECT_RELATION),
            (ForeignObjectRel, constants.CT_REVERSE_RELATION),
        )

        for field_base_class, column_type in types_map:
            if isinstance(self.field, field_base_class):
                return column_type

        return constants.CT_OTHER

    @property
    def choices(self):
        u"""Допустимые значения столбца с описанием."""
        return self.field.choices or None

    @property
    def title(self):
        if isinstance(self.field, ForeignObjectRel):
            options = getattr(self.field.related_model, '_meta')
            if isinstance(self.field, OneToOneRel):
                result = options.verbose_name
            else:
                result = options.verbose_name_plural
        else:
            result = self.field.verbose_name

        return force_unicode(result)

    def get_full_title(self, delimiter=u' → '):
        u"""Возвращает полное наименование с учетом иерархии.

        :param unicode delimiter: Разделитель наименований.

        :rtype: unicode
        """
        if self.parent:
            return delimiter.join((
                self.parent.get_full_title(delimiter), self.title
            ))
        else:
            return self.title

    def is_root(self):
        return not self.parent

    @property
    def root(self):
        result = self

        while result.parent:
            result = result.parent

        return result

    def has_nested_columns(self):
        if isinstance(self.field, ForeignObjectRel):
            model = self.field.related_model
        elif isinstance(self.field, RelatedField):
            model = self.field.rel.to
        else:
            return False

        for field in getattr(model, '_meta').get_fields():
            if not skip_field(model, field):
                return True

        return False

    def get_nested_columns(self):
        if isinstance(self.field, ForeignObjectRel):
            model = self.field.related_model
        elif isinstance(self.field, RelatedField):
            model = self.field.rel.to
        else:
            return ()

        return (
            ColumnDescriptor(self, field, level=self.level + 1)
            for field in getattr(model, '_meta').get_fields()
            if not skip_field(model, field)
        )

    def validate_value(self, value):
        u"""Проверяет текстовое представление значения на соотв-е типу поля.

        :param unicode value: Текстовое представление значения для проверки.

        :raises django.core.exceptions.ValidationError: если значение аргумента
            value не прошло проверку. Описание ошибки будет в исключении.
        """
        if self.data_type == constants.CT_REVERSE_RELATION:
            raise ValidationError(u'Обратная связь не доступна для сравнения.')
        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

        elif self.data_type == constants.CT_DIRECT_RELATION:
            raise ValidationError(u'Ключевые поля не доступны для сравнения.')
        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

        elif self.data_type == constants.CT_CHOICES:
            lower_case_value = value.lower()
            for _, choice_display in self.field.choices:
                if choice_display.lower() == lower_case_value:
                    break
            else:
                raise ValidationError(
                    u'Недопустимое значение для сравнения: {}.'
                    .format(value)
                )
        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

        else:
            form_field = self.field.formfield()
            if form_field:
                form_field.validate(value)
        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -


class ModelDataSourceParams(object):
    u"""Параметры для дескриптора источника данных в шаблоне отчета."""

    def __init__(self, model, filter_func=None):
        self._model = LazyModel(model)
        self._filter_func = filter_func

    @cached_property
    def name(self):
        u"""Внутрисистемное имя источника данных."""
        result = '.'.join((self._model.app_label, self._model.model_name))

        return force_unicode(result)

    @cached_property
    def title(self):
        u"""Видимое пользователем наименование ИД."""
        model = self._model.get_model()
        result = getattr(model, '_meta').verbose_name
        return force_unicode(result)

    def get_data_source_descriptor(self):
        u"""Возвращает дескриптор источника данных."""
        return ModelDataSourceDescriptor(
            self._model.get_model(), self._filter_func
        )


class ModelDataSourceDescriptor(object):

    u"""Дескриптор модели, являющейся источником данных в шаблоне отчета."""

    def __init__(self, model, filter_func=None):
        self._model = LazyModel(model)
        self.column_descriptors_cache = {}
        self._filter_func = filter_func

    @cached_property
    def model(self):
        return self._model.get_model()

    @cached_property
    def name(self):
        u"""Внутрисистемное имя источника данных."""
        result = '.'.join((self._model.app_label, self._model.model_name))

        return force_unicode(result)

    @cached_property
    def title(self):
        u"""Видимое пользователем наименование ИД."""
        model = self._model.get_model()
        result = getattr(model, '_meta').verbose_name
        return force_unicode(result)

    @cached_property
    def _options(self):
        return ModelOptions(self._model.get_model())

    def add_source_filter(self, query, include_available_units, user):
        u"""Добавление фильтра к источнику данных."""
        if self._filter_func:
            return self._filter_func(query, include_available_units, user)
        return query

    def get_available_columns(self):
        model = self._model.get_model()
        return (
            ColumnDescriptor(None, field)
            for field in getattr(model, '_meta').get_fields()
            if not skip_field(model, field)
        )

    def get_column_descriptor(self, accessor_name):
        u"""Возвращает дескриптор столбца по его имени.

        .. code-block:: python

           data_source.get_column_descriptor('person.surname')

        :rtype: ColumnDescriptor
        """
        level_names = []
        model = self.model
        result = None
        for level_name in accessor_name.split('.'):
            level_names.append(level_name)
            full_accessor_name = '.'.join(level_names)

            if not model:
                raise FieldDoesNotExist(full_accessor_name)
            # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
            # Поиск поля в модели.

            for field in getattr(model, '_meta').get_fields():
                if skip_field(model, field):
                    continue

                if isinstance(field, ForeignObjectRel):
                    field_name = field.get_accessor_name()
                    field_model = field.related_model
                else:
                    field_name = field.name
                    if isinstance(field, RelatedField):
                        field_model = field.rel.to
                    else:
                        field_model = None

                if field_name == level_name:
                    column_field = field
                    model = field_model
                    break
            else:
                raise FieldDoesNotExist(
                    '{}.{}'.format(model.__name__, full_accessor_name)
                )
            # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
            # Создание дескриптора столбца, если его нет в кэше.

            if full_accessor_name in self.column_descriptors_cache:
                result = self.column_descriptors_cache[full_accessor_name]
            else:
                result = ColumnDescriptor(
                    result, column_field, len(level_names) - 1
                )
                self.column_descriptors_cache[full_accessor_name] = result
            # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

        return result
